from scipy import io
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.metrics import precision_recall_fscore_support, accuracy_score
import argparse
from distutils.util import strtobool
from vae import VAE
import tensorflow as tf
import os,sys
import plots
import pandas as pd

def main(args):

    fdir = '%s_%s_%s/%s/ch%s_eh%s_q%.1f_l1_%s_l2_%s_lr%s_%s/take%s' % (args.checkpoint_dir, args.model, args.normalize,
                                                                    args.dataset,
                                                            args.comp_hiddens, args.est_hiddens, args.qstep, args.lambda1,
                                                            args.lambda2, args.learning_rate, args.loss, args.take)
    print(fdir)
    if not os.path.exists(fdir):
        os.makedirs(fdir)

    if args.dataset != 'toy':
        if args.dataset == 'Arrhythmia':
            matdata = io.loadmat('dataset/arrhythmia.mat', squeeze_me=True)
           #print(matdata)
            X = np.array(matdata['X'])
            y = np.array(matdata['y'], dtype='int32')
            #X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.50, random_state=123)
            #X_train, y_train = X_train[y_train == 0], y_train[y_train == 0]
            log_step = 100
        elif args.dataset == 'Thyroid':
            matdata = io.loadmat('dataset/thyroid.mat', squeeze_me=True)
           #print(matdata)
            X = np.array(matdata['X'])
            y = np.array(matdata['y'], dtype='int32')
            #X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.50, random_state=123)
            #X_train, y_train = X_train[y_train == 0], y_train[y_train == 0]
            log_step = 100

        elif args.dataset == 'KDDCup99':
            url_base = "http://kdd.ics.uci.edu/databases/kddcup99"
            # KDDCup 10% Data
            # url_data = f"{url_base}/kddcup.data_10_percent.gz"
            url_data = "dataset/kddcup.data_10_percent.gz"
            # info data (column names, col types)
            # url_info = f"{url_base}/kddcup.names"
            url_info = "dataset/kddcup.names.csv"
            # In[3]:
            # Import info data
            df_info = pd.read_csv(url_info, sep=":", skiprows=1, index_col=False, names=["colname", "type"])
            colnames = df_info.colname.values
            coltypes = np.where(df_info["type"].str.contains("continuous"), "float", "str")
            colnames = np.append(colnames, ["status"])
            coltypes = np.append(coltypes, ["str"])
            # Import data
            df = pd.read_csv(url_data, names=colnames, index_col=False,
                             dtype=dict(zip(colnames, coltypes)))
            # In[4]:
            # Dumminize
            X = pd.get_dummies(df.iloc[:, :-1]).values

            # Create Traget Flag
            # Anomaly data when status is normal, Otherwise, Not anomaly.
            y = np.where(df.status == "normal.", 1, 0)
            #X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.50, random_state=123)
            #X_train, y_train = X_train[y_train == 0], y_train[y_train == 0]
            log_step = 1
        elif args.dataset == 'KDDCup-rev':
            url_base = "http://kdd.ics.uci.edu/databases/kddcup99"
            url_data = "dataset/kddcup.data_10_percent.gz"
            url_info = "dataset/kddcup.names.csv"
            df_info = pd.read_csv(url_info, sep=":", skiprows=1, index_col=False, names=["colname", "type"])
            colnames = df_info.colname.values
            coltypes = np.where(df_info["type"].str.contains("continuous"), "float", "str")
            colnames = np.append(colnames, ["status"])
            coltypes = np.append(coltypes, ["str"])
            df = pd.read_csv(url_data, names=colnames, index_col=False,
                             dtype=dict(zip(colnames, coltypes)))
            X = pd.get_dummies(df.iloc[:, :-1]).values
            y = np.where(df.status == "normal.", 0, 1)


            if args.normalize == 'min-max':
                from sklearn.preprocessing import MinMaxScaler
                mm = MinMaxScaler()
                X = mm.fit_transform(X)
                normalize = False
            else:
                normalize = True


            #normal data
            normal_X, normal_y = X[y==0],y[y==0]
            #randIdx = np.arange(normal_X.shape[0])
            randIdx = np.load('dataset/%s/normal_take%s.npy' % (args.dataset, args.take))
            normal_X, normal_y = normal_X[randIdx], normal_y[randIdx]
            #anomaly
            attack_X, attack_y = X[y==1],y[y==1]
            N_attack = attack_X.shape[0]
            #Get random attack data
            #randIdx = np.arange(N_attack)
            #np.random.shuffle(randIdx)
            randIdx = np.load('dataset/%s/attack_take%s.npy' % (args.dataset, args.take))
            #extract attack data
            att_num = normal_X.shape[0]//4
            train_att_x, train_att_y = attack_X[randIdx[:att_num]], attack_y[randIdx[:att_num]]

            split_attack = train_att_x.shape[0]//2
            split_norm = normal_X.shape[0]//2

            # train dataset: only normal data
            #X_train_att = train_att_x[:split_attack]
            #y_train_att = train_att_y[:split_attack]
            X_train = normal_X[:split_norm]
            y_train = normal_y[:split_norm]

            # test dataset: both normal and attack
            X_test_att = train_att_x[:split_attack]
            y_test_att = train_att_y[:split_attack]
            X_test_norm = normal_X[split_norm:]
            y_test_norm = normal_y[split_norm:]
            X_test = np.concatenate((X_test_att, X_test_norm), axis=0)
            y_test = np.concatenate((y_test_att, y_test_norm), axis=0)

            print('KDDCup-rev train shape:', X_train.shape, 'Test shape:', X_test.shape)
            #network parameters
            log_step = 1
            #args.comp_hiddens = '60_30_10_1'
            #args.est_hiddens = '10_2'
            #args.epoch_size = 100
            #args.batch_size = 1024

            '''
            if args.normalize == 'min-max':
                from sklearn.preprocessing import MinMaxScaler
                mm = MinMaxScaler()
                X_train = mm.fit_transform(X_train)
                X_test = mm.fit_transform(X_test)

                normalize = False
            else:
                normalize = True
            '''

        else:
            print('ERROR %s is invalid dataset'%(args.dataset))
            return -1

        #test = np.array([[1,2,3],
        #                 [2,3,4]])
        #print('test',test, test+1)

        if args.normalize == 'min-max' and args.dataset != 'KDDCup-rev':
            '''
            X_min = X.min()
            X_max = X.max()
            X = (X - X_min) / (X_max - X_min)
            print('X_min',X_min, 'X_Max',X_max)
            '''
            from sklearn.preprocessing import MinMaxScaler
            mm = MinMaxScaler()
            X = mm.fit_transform(X)

            normalize = False
        else:
            normalize = True

        if args.dataset != 'KDDCup-rev':
            normal_data = X[y == 1]
            normal_labels = y[y == 1]

            N_normal = normal_data.shape[0]

            attack_data = X[y == 0]
            attack_labels = y[y == 0]

            N_attack = attack_data.shape[0]


            randIdx = np.load('dataset/%s/train_take%s.npy'%(args.dataset, args.take))
            randIdx_test = np.load('dataset/%s/test_take%s.npy'%(args.dataset, args.take))
            print(randIdx[:10])

            N_train = N_attack // 2

            X_train = attack_data[randIdx[:N_train]]
            y_train = attack_labels[randIdx[:N_train]]

            X_test = attack_data[randIdx[N_train:]]
            y_test = attack_labels[randIdx[N_train:]]

            X_test = np.concatenate((X_test, normal_data[randIdx_test[:int(N_normal / 2)], :]), axis=0)
            y_test = np.concatenate((y_test, normal_labels[randIdx_test[:int(N_normal / 2)]]), axis=0)



    print(X.shape, y.shape)
    # Split Data

    print('X',X.shape)
    print('y',y.shape)
    print('train',X_train.shape)
    print('test',X_test.shape)

    s_comp_h = args.comp_hiddens.split('_')
    comp_hiddens = []
    for s in s_comp_h:
        comp_hiddens.append(int(s))
    print(comp_hiddens)

    s_comp_e = args.est_hiddens.split('_')
    est_hiddens = []
    for s in s_comp_e:
        est_hiddens.append(int(s))
    print(est_hiddens)

    args.fdir = fdir
    score_dir = '%s_%s_%s/%s'%(args.checkpoint_dir, args.model, args.normalize, args.dataset)
    args.score_dir = score_dir
    args.score_fname = '%s_%s.csv'% (args.model, args.dataset)
    args.score_fname_b = '%s_%s_best.csv'% (args.model, args.dataset)
    args.score_fname_f = '%s_%s_final.csv' % (args.model, args.dataset)
    args.params = 'ch%s_eh%s_q%.1f_l1_%s_l2_%s_lr%s_%s/take%s' % (args.comp_hiddens, args.est_hiddens, args.qstep, args.lambda1,
                                                            args.lambda2, args.learning_rate, args.loss, args.take)
    args.log_step = log_step

    if args.ac == 'leaky_relu':
        activation = tf.nn.leaky_relu
    else:
        activation = tf.nn.tanh
    model = VAE(args=vars(args),
        comp_hiddens=comp_hiddens, comp_activation=activation,
        est_hiddens=est_hiddens, est_dropout_ratio=0.5, est_activation=tf.nn.tanh,
        random_seed=1111, normalize=normalize
    )

    # In[7]:
    if args.mode == 'train':
        model.fit(X_train, X_test, y_test)
        model.save(fdir,'final')
    elif args.mode == 'test':
        model.restore(fdir,'final')
        model.test(X_train)
    # ## Apply model to test data



if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--mode', choices=["train", 'test', 'plot'],
        help="What to do: 'train' loads training data and trains (or continues "
        "to train) a new model. 'test' load trained model and test.")
    parser.add_argument(
        "--model", default="DAGMM",
        help="name of model")
    parser.add_argument(
        "--dataset", required=True, default="Toy",
        help="name of dataset(Arrhythmia, KDDCup99, KDDCup-rev) ")
    parser.add_argument(
        "--batch_size", type=int, default=128,
        help="Batch size for training.")
    parser.add_argument(
        "--learning_rate", type=float, default=1e-4,
        help="learning rate")
    parser.add_argument(
        "--epoch_size", type=int, default=10000,
        help="Train up to this number of epochs.")
    parser.add_argument(
        "--display_step", type=int, default=2000,
        help="save loss for plot every this number of steps.")
    parser.add_argument(
        "--lambda1", type=float, default=1,
        help="Lambda for rate-distortion tradeoff.")
    parser.add_argument(
        "--lambda2", type=float, default=1000,
        help="Lambda for rate-distortion tradeoff.")
    parser.add_argument(
        "--qstep", type=float, default=0.1,
        help="quantization step")
    parser.add_argument(
        "--checkpoint_dir", default="cache",
        help="Directory where to save/load model checkpoints.")
    parser.add_argument(
        "--stds", default="std\\std_frdc_loss_lambda10",
        help="Filename to save std log")
    parser.add_argument(
        "--comp_hiddens", default="10_2",
        help="number of compression layers")
    parser.add_argument(
        "--est_hiddens", default="10_2",
        help="number of hidden layers")
    parser.add_argument(
        "--take", default="0",
        help="number_of_take")
    parser.add_argument(
        "--normalize", default="min-max",
        help="normalize")
    parser.add_argument(
        "--zr", type=strtobool, default=True,
        help="use zr or not")
    parser.add_argument(
        "--loss", default='SSE',
        help="use zr or not")
    parser.add_argument(
        "--energy", default='pdf',
        help="use zr or not")

    parser.add_argument(
        "--ac", default='tanh',
        help="use zr or not")

    parser.add_argument('--model_save_step', type=int, default=1000)

    parser.add_argument('-gpu','--gpu_id',
        help='GPU device id to use [0]',default=0, type=int)

    args = parser.parse_args()
    main(args=args)
