import tensorflow as tf
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.externals import joblib
from vae.compression_net import CompressionNet


from os import makedirs
from os.path import exists, join
import os, sys
import scipy.io as sio

from tensorflow.python.ops import math_ops
from tensorflow.python.ops import random_ops
from tensorflow.python.ops import array_ops
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import dtypes
from tensorflow.python.keras.engine import base_layer

import plots

from sklearn.metrics import precision_recall_fscore_support, accuracy_score
from scipy import io, stats

class VAE:
    """ Deep Autoencoding Gaussian Mixture Model.

    This implementation is based on the paper:
    Bo Zong+ (2018) Deep Autoencoding Gaussian Mixture Model
    for Unsupervised Anomaly Detection, ICLR 2018
    (this is UNOFFICIAL implementation)
    """
    DEFAULTS = {}
    MODEL_FILENAME = "VAE_model"
    SCALER_FILENAME = "VAE_scaler"

    def __init__(self, args, comp_hiddens, comp_activation, est_hiddens, est_activation, est_dropout_ratio=0.5,
                 normalize=False, random_seed=123):

        self.__dict__.update(VAE.DEFAULTS, **args)
        #
        self.comp_net = CompressionNet(comp_hiddens, comp_activation)
        self.est_dropout_ratio = est_dropout_ratio

        self.n_comp = est_hiddens[-1]
        self.comp_hiddens = comp_hiddens

        self.normalize = normalize
        self.scaler = None
        self.seed = random_seed

        self.graph = None
        self.sess = None

        self.best_fscore = 0
        self.best_acc = 0
        self.best_prec = 0
        self.best_recall = 0
        self.best_epoch = 0

    def __del__(self):
        if self.sess is not None:
            self.sess.close()

    def fit(self, x, x_test, y_test):
        """ Fit the DAGMM model according to the given data.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            Training data.
        """
        n_samples, n_features = x.shape

        if self.normalize:
            self.scaler = scaler = StandardScaler()
            x = scaler.fit_transform(x)
            x_test = scaler.transform(x_test)

        with tf.Graph().as_default() as graph:
            self.graph = graph
            tf.set_random_seed(self.seed)
            np.random.seed(seed=self.seed)

            # Create Placeholder
            self.input = input = tf.placeholder(
                dtype=tf.float32, shape=[None, n_features])
            self.drop = drop = tf.placeholder(dtype=tf.float32, shape=[])

            mean, sigma, z = self.comp_net.compress(input)
            #x_dash, tensor1 = self.comp_net.reverse(mean)
            x_dash = self.comp_net.reverse(mean)

            x_dash2 = self.comp_net.reverse(z)


            if self.loss == 'log':
                loss_r = tf.log(tf.clip_by_value(self.comp_net.reconstruction_error(input, x_dash), 1e-6, 1e10))
            elif self.loss == 'L2':
                loss_r = self.comp_net.reconstruction_errorL2(input, x_dash)
            else:
                loss_r = self.comp_net.reconstruction_error(input, x_dash)

            loss_range = self.comp_net.reconstruction_error(x_dash, x_dash2)

            kl_div_loss = 1 + sigma - tf.square(mean) - tf.exp(sigma)
            kl_div_loss = -0.5 * tf.reduce_sum(kl_div_loss, 1)
            kl_div_loss_s = kl_div_loss
            #print('kl_shape', kl_div_loss.shape)
            kl_div_loss = tf.reduce_mean(kl_div_loss)

            #entropy = tf.reduce_mean(energy)
            entropy = kl_div_loss
            loss = entropy + self.lambda1 * loss_r + self.lambda2 * loss_range

            #Calcurate reconstruction loss average
            x_dashp = self.comp_net.reverse(mean + tf.exp(sigma/2))
            x_dashm = self.comp_net.reverse(mean - tf.exp(sigma/2))
            loss_avg = 0.5 * (tf.reduce_sum(tf.square(input - x_dashp), axis=1) + tf.reduce_sum(tf.square(input - x_dashm), axis=1))

            # Minimizer
            minimizer = tf.train.AdamOptimizer(self.learning_rate).minimize(loss)
            self.saver = tf.train.Saver()

            # Number of batch
            n_batch = (n_samples - 1) // self.batch_size + 1

            # Create tensorflow session and initilize
            init = tf.global_variables_initializer()

            self.sess = tf.Session(graph=graph)
            self.sess.run(init)

            # Training
            idx = np.arange(x.shape[0])
            np.random.shuffle(idx)

            log_name = os.path.join(self.fdir, 'train.log')
            logs = open(log_name, 'w')

            saver_step = self.epoch_size // 5
            if self.model_save_step > self.epoch_size:
                self.model_save_step = self.epoch_size // 5
            print('model_save_step', self.model_save_step)

            for epoch in range(self.epoch_size):
                for batch in range(n_batch):
                    i_start = batch * self.batch_size
                    i_end = (batch + 1) * self.batch_size
                    x_batch = x[idx[i_start:i_end]]

                    self.sess.run(minimizer, feed_dict={
                        input: x_batch, drop: self.est_dropout_ratio})

                if (epoch + 1) % self.log_step == 0:
                    loss_val, lr, e, lr2, kl_div_loss_s_ = self.sess.run(
                        [loss, loss_r, entropy, loss_range, kl_div_loss_s],
                        feed_dict={input: x, drop: 0})
                    vals = " epoch,{},{},loss,{:.3f},rate,{:.3f},loss_rec,{:.8f},loss_range,{:.8f}".format(epoch + 1,
                                                                                                           self.epoch_size,
                                                                                                           loss_val, e,
                                                                                                          lr, lr2)
                    print(vals)
                    logs.write(vals + '\n')
                    # save GMM parameters

                if (epoch + 1) % self.model_save_step == 0:
                    if 1:  # add
                        # print('Training GMM parameters in each model_save_step')
                        print(self.loss)
                        z_, mean_, sigma_, kl_div_loss_s_, loss_avg_ = self.sess.run([z, mean, sigma, kl_div_loss_s, loss_avg], feed_dict={input: x_test, drop: 0})
                        if self.energy == 'kld':
                            energy_t = kl_div_loss_s_
                        elif self.energy == 'L':
                            energy_t = kl_div_loss_s_ + self.lambda2 * loss_avg_
                        else:
                            pdfs = []
                            pdfs_mode = []

                            for i in range(z_.shape[0]):
                                # print(i)
                                pdf = 0
                                pdf_m = 0
                                for d in range(z_.shape[1]):
                                    pdf += np.log(stats.norm.pdf(mean_[i][d], loc=0, scale=1))
                                    pdf_m += np.log(stats.norm.pdf(mean_[i][d], loc=0, scale=1) * np.exp(sigma_[i][d] / 2))
                                pdfs.append(np.exp(pdf))
                                pdfs_mode.append(np.exp(pdf_m))

                            energy_t = np.array(-np.log(pdfs_mode))

                        if self.dataset == 'KDDCup99' or self.dataset == 'KDDCup-rev':
                            anomaly_energy_threshold = np.percentile(energy_t, 80)
                        elif self.dataset == 'Arrhythmia':
                            anomaly_energy_threshold = np.percentile(energy_t, 85)
                        elif self.dataset == 'Thyroid':
                            anomaly_energy_threshold = np.percentile(energy_t, 97.5)
                        # anomaly_energy_threshold = np.percentile(energy_t, 80)
                        print(' anomaly_energy_threshold with GMM parameters', anomaly_energy_threshold)
                        y_pred_flag = np.where(energy_t >= anomaly_energy_threshold, 1, 0)
                        precision, recall, f_score, _ = precision_recall_fscore_support(y_test, y_pred_flag,
                                                                                        average="binary")
                        accuracy = accuracy_score(y_test, y_pred_flag)

                        #self.save(self.fdir, str(epoch + 1))
                    else:
                        self.z = z
                        self.energy = self.gmm.energy(z)
                        print('No No No .Fix GMM in each model_save_step')
                        accuracy, precision, recall, f_score = self.calc_result(x_test, y_test)

                    # accuracy, precision, recall, f_score = self.calc_result(x_test, y_test)
                    #score_path = os.path.join(self.score_dir, self.score_fname)
                    #f = open(score_path, 'a')
                    val = '_%d, accuracy,%f, precision,%f, recall,%f, f_score,%f,' % (epoch + 1,
                                                                                      accuracy, precision, recall,
                                                                                      f_score)
                    #f.write(self.params + val + '\n')
                    print(val)
                    #f.close()

                    if f_score > self.best_fscore:
                        self.best_fscore = f_score
                        self.best_acc = accuracy
                        self.best_prec = precision
                        self.best_recall = recall
                        self.best_epoch = epoch + 1

            score_path = os.path.join(self.score_dir, self.score_fname_b)
            f = open(score_path, 'a')
            if self.take == "0":
                f.write('\n')
            val = '%s_%d, accuracy,%f, precision,%f, recall,%f, f_score,%f,\n' % (self.params, self.best_epoch,
                                                                                  self.best_acc, self.best_prec,
                                                                                  self.best_recall,
                                                                                  self.best_fscore)
            f.write(val)
            f.close()
            #self.saver.save(self.sess, self.fdir)

    def energy_for_test(self, z_test, phi_test, mu_test, L_test):
        """ calculate an energy of each row of z

        Parameters
        ----------
        z : tf.Tensor, shape (n_samples, n_features)
            data each row of which is calculated its energy.

        Returns
        -------
        energy : tf.Tensor, shape (n_samples)
            calculated energies
        """

        with tf.variable_scope("Energy_test"):
            # Instead of inverse covariance matrix, exploit cholesky decomposition
            # for stability of calculation.
            z_centered = z_test[:, None, :] - mu_test[None, :, :]  # ikl
            v = tf.matrix_triangular_solve(L_test, tf.transpose(z_centered, [1, 2, 0]))  # kli

            # log(det(Sigma)) = 2 * sum[log(diag(L))]
            log_det_sigma = 2.0 * tf.reduce_sum(tf.log(tf.matrix_diag_part(L_test)), axis=1)

            # To calculate energies, use "log-sum-exp" (different from orginal paper)
            d = z_test.get_shape().as_list()[1]
            logits = tf.log(phi_test[:, None]) - 0.5 * (tf.reduce_sum(tf.square(v), axis=1)
                                                        + d * tf.log(2.0 * np.pi) + log_det_sigma[:, None])
            energies = - tf.reduce_logsumexp(logits, axis=0)

        return energies

    def get_variable_via_scope(self, scope_lst):
        vars = []
        for sc in scope_lst:
            sc_variable = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=sc)
            vars.extend(sc_variable)
        return vars

    def predict(self, x):
        """ Calculate anormaly scores (sample energy) on samples in X.

        Parameters
        ----------
        x : array-like, shape (n_samples, n_features)
            Data for which anomaly scores are calculated.
            n_features must be equal to n_features of the fitted data.

        Returns
        -------
        energies : array-like, shape (n_samples)
            Calculated sample energies.
        """
        if self.sess is None:
            raise Exception("Trained model does not exist.")

        if self.normalize:
            x = self.scaler.transform(x)

        energies, z = self.sess.run([self.energy, self.z], feed_dict={self.input: x})
        return energies, z

    def calc_result(self, x, y):
        y_pred, z = self.predict(x)

        if self.dataset == 'KDDCup99' or self.dataset == 'KDDCup-rev':
            anomaly_energy_threshold = np.percentile(y_pred, 80)
        elif self.dataset == 'Arrhythmia':
            anomaly_energy_threshold = np.percentile(y_pred, 85)
        elif self.dataset == 'Thyroid':
            anomaly_energy_threshold = np.percentile(y_pred, 97.5)
        # print(f"Energy thleshold to detect anomaly : {anomaly_energy_threshold:.3f}")

        y_pred_flag = np.where(y_pred >= anomaly_energy_threshold, 1, 0)
        prec, recall, fscore, _ = precision_recall_fscore_support(y, y_pred_flag, average="binary")
        accuracy = accuracy_score(y, y_pred_flag)
        # precision, recall, f_score, support = prf(gt, pred, average='binary')
        # print("Accuracy : {:0.4f}, Precision : {:0.4f}, Recall : {:0.4f}, F-score : {:0.4f}".format(accuracy, prec, recall, fscore))
        return accuracy, prec, recall, fscore

    def save(self, fdir, epoch):
        """ Save trained model to designated directory.
        This method have to be called after training.
        (If not, throw an exception)
WA
        Parameters
        ----------
        fdir : str
            Path of directory trained model is saved.
            If not exists, it is created automatically.
        """
        if self.sess is None:
            raise Exception("Trained model does not exist.")

        if not exists(fdir):
            makedirs(fdir)

        model_path = join(fdir, self.MODEL_FILENAME + epoch)
        self.saver.save(self.sess, model_path)

        if self.normalize:
            scaler_path = join(fdir, self.SCALER_FILENAME)
            joblib.dump(self.scaler, scaler_path)

    def restore(self, fdir):
        """ Restore trained model from designated directory.

        Parameters
        ----------
        fdir : str
            Path of directory trained model is saved.
        """
        if not exists(fdir):
            raise Exception("Model directory does not exist.")

        model_path = join(fdir, self.MODEL_FILENAME)
        meta_path = model_path + ".meta"

        with tf.Graph().as_default() as graph:
            self.graph = graph
            self.sess = tf.Session(graph=graph)
            self.saver = tf.train.import_meta_graph(meta_path)
            self.saver.restore(self.sess, model_path)

            self.input, self.energy = tf.get_collection("save")

        if self.normalize:
            scaler_path = join(fdir, self.SCALER_FILENAME)
            self.scaler = joblib.load(scaler_path)

    def test(self, X, x, model_path):
        n_samples, n_features = x.shape

        '''if self.normalize:
            self.scaler = scaler = StandardScaler()
            x = scaler.fit_transform(x)'''
        if self.normalize:
            scaler = StandardScaler()
            X = scaler.fit_transform(X)
            x = scaler.transform(x)
            print('Mean of x_test in predict:', np.mean(x))

        # Create Placeholder
        self.input = input = tf.placeholder(
            dtype=tf.float32, shape=[None, n_features])
        self.drop = drop = tf.placeholder(dtype=tf.float32, shape=[])

        # Build graph
        # z_c, z, x_dash,tensor  = self.comp_net.inference(input)
        z_c = self.comp_net.compress(input)
        x_dash, tensor = self.comp_net.reverse(z_c)
        z = self.comp_net.extract_feature(input, x_dash, z_c)
        gamma = self.est_net.inference(z, self.drop)
        phi, mean, sigma, L = self.gmm.fit(z, gamma)
        energy = self.gmm.energy(z)

        # assign
        p_z = tf.placeholder(dtype=tf.float32, shape=[None, 3])
        p_phi = tf.placeholder(dtype=tf.float32, shape=[self.n_comp])
        p_mean = tf.placeholder(dtype=tf.float32, shape=[self.n_comp, 3])
        p_L = tf.placeholder(dtype=tf.float32, shape=[self.n_comp, 3, 3])
        energy_test = self.energy_for_test(p_z, p_phi, p_mean, p_L)

        # restore model
        with tf.Session() as sess:
            latest = tf.train.latest_checkpoint(checkpoint_dir=model_path)
            tf.train.Saver().restore(sess, save_path=latest)
            phi_tr, mean_tr, L_tr = sess.run([phi, mean, L], feed_dict={input: X, drop: 0})
            print('test mode:', phi_tr)
            z_te = sess.run(z, feed_dict={input: x, drop: 0})
            energy_te = sess.run(energy_test, feed_dict={p_z: z_te, p_phi: phi_tr, p_mean: mean_tr, p_L: L_tr})
            return energy_te

